"""Value model"""

from typing import Any, Union
from pydantic import BaseModel


class Value(BaseModel):
    """Value model"""

    value: Union[Any, None]

    class Config:
        """Config"""

        arbitrary_types_allowed = True


__all__ = ("Value",)



""".. include:: ../README.md"""

from .models import Connection, Empty, ResultModel, Value
from .coleridge import Coleridge
from .decorator import ColeridgeDecorator
from .decorated import DecoratedBackgroundFunction
from .rabbit import RabbitBackgroundFunction

__all__ = (
    "Coleridge",
    "ColeridgeDecorator",
    "DecoratedBackgroundFunction",
    "RabbitBackgroundFunction",
    "Connection",
    "Empty",
    "ResultModel",
    "Value",
)



"""Decorator utils"""

from pathlib import Path
from typing import Generic, TypeVar, Type, Union, List, Callable, Literal
from pydantic import BaseModel
from .decorated import DecoratedBackgroundFunction
from .models.connection import Connection
from .rabbit import RabbitBackgroundFunction

T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)


class ColeridgeDecorator(Generic[T, U]):
    """Decorator utils"""

    _input_type: Type[T]
    _output_type: Type[U]
    _on_finish: Union[Callable[[Union[U, List[U]]], None], None]
    _on_error: Union[Callable[[Exception], None], None]
    _on_finish_signal: Union[Callable[[], None], None]
    _mode: Literal["rabbit", "background"]
    _connection_settings: Union[Connection, None, str, Path]
    _queue: Union[str, None]

    def __init__(  # noqa: PLR0913
        self,
        input_type: Type[T],
        output_type: Type[U],
        *,
        mode: Literal["rabbit", "background"] = "background",
        connection_settings: Union[Connection, None, str, Path] = None,
        queue: Union[str, None] = None,
        on_finish: Union[Callable[[Union[U, List[U]]], None], None] = None,
        on_error: Union[Callable[[Exception], None], None] = None,
        on_finish_signal: Union[Callable[[], None], None] = None,
    ) -> None:
        """
        Initializes a new instance of the ColeridgeDecorator class.

        Args:
            input_type (Type[T]): The type of the input data.
            output_type (Type[U]): The type of the output data.
            mode (Literal["rabbit", "background"], optional): The mode \
                of the decorator. Defaults to "background".
            connection_settings (Union[Connection, None, str, Path], optional): The connection \
                settings. Defaults to None.
            queue (Union[str, None], optional): The queue name. Defaults to None.
            on_finish (Union[Callable[[Union[U, List[U]]], None], None], optional): The \
                callback function to call when the task is finished. Defaults to None.
            on_error (Union[Callable[[Exception], None], None], optional): The callback \
              function to call when an error occurs. Defaults to None.
            on_finish_signal (Union[Callable[[], None], None], optional): The callback \
            function to call when the task is finished with a signal. Defaults to None.

        Returns:
            None
        """
        self._input_type = input_type
        self._output_type = output_type
        self._on_finish = on_finish
        self._on_error = on_error
        self._on_finish_signal = on_finish_signal

        self._mode = mode
        self._connection_settings = connection_settings
        self._queue = queue

    def __call__(
        self,
        func: Callable[[Union[T, List[T]]], Union[U, List[U]]],
    ) -> Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]:
        """
        Calls the decorated function with the provided parameters.

        Args:
            func (Callable[[Union[T, List[T]]], Union[U, List[U]]]): The function to be called.

        Returns:
            Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]: 
            The result of the function call, which can be either a \
            DecoratedBackgroundFunction or a RabbitBackgroundFunction.
        """
        if self._mode == "rabbit":
            rabbit: RabbitBackgroundFunction[T, U] = RabbitBackgroundFunction(
                func,
                self._input_type,
                self._output_type,
                connection_settings=self._connection_settings,
                queue=self._queue,
            )
            if self._on_finish is not None:
                rabbit.on_finish = self._on_finish
            if self._on_error is not None:
                rabbit.on_error = self._on_error
            if self._on_finish_signal is not None:
                rabbit.on_finish_signal = self._on_finish_signal
            return rabbit
        dec: DecoratedBackgroundFunction[T, U] = DecoratedBackgroundFunction(
            func,
            self._input_type,
            self._output_type,
        )
        if self._on_finish is not None:
            dec.on_finish = self._on_finish
        if self._on_error is not None:
            dec.on_error = self._on_error
        if self._on_finish_signal is not None:
            dec.on_finish_signal = self._on_finish_signal
        return dec


__all__ = (
    "ColeridgeDecorator",
    "T",
    "U",
)



"""Models for Coleridge."""

from .connection import Connection
from .empty import Empty
from .response import ResultModel
from .value import Value

__all__ = (
    "Connection",
    "Empty",
    "ResultModel",
    "Value",
)



"""Module for RabbitMQ utilities
"""

from pathlib import Path
from datetime import datetime
from typing import (
    AnyStr,
    List,
    Union,
    Any,
    Callable,
    Type,
    Generic,
    TypeVar,
    Dict,
    cast,
)
from time import sleep
from uuid import uuid4
from pickle import dumps, loads  # nosec B403
from json import loads as json_loads
from threading import Thread
from yaml import load, SafeLoader
from pika import BlockingConnection, ConnectionParameters, PlainCredentials
from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import AMQPConnectionError
from pydantic import BaseModel
from .models.connection import Connection
from .models.response import ResultModel
from .result import ExecutionResult as Result

T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)


class RabbitBackgroundFunction(Generic[T, U]):
    """Background function using RabbitMQ"""

    _data: Dict[str, ResultModel[U]]
    _input_type: Type[T]
    _output_type: Type[U]
    _on_finish: Callable[[Union[U, List[U]]], None]
    _on_error: Callable[[Exception], None]
    _on_finish_signal: Callable[[], None]
    func: Callable[[Union[T, List[T]]], Union[U, List[U]]]
    _queue: str
    _client: BlockingConnection
    _channel: BlockingChannel

    def __init__(  # noqa: C901, PLR0912, PLR0915
        self,
        func: Callable[[Union[T, List[T]]], Union[U, List[U]]],
        input_type: Type[T],
        output_type: Type[U],
        connection_settings: Union[Connection, None, str, Path] = None,
        queue: Union[str, None] = None,
    ) -> None:
        """
        Initializes a new instance of the RabbitBackgroundFunction class.

        Args:
            func (Callable[[Union[T, List[T]]], Union[U, List[U]]]): The function to be \
                executed in the background.
            input_type (Type[T]): The type of the input data.
            output_type (Type[U]): The type of the output data.
            connection_settings (Union[Connection, None, str, Path], optional): The connection \
                settings for the RabbitMQ server. Defaults to None.
            queue (Union[str, None], optional): The name of the queue to use. Defaults to None.

        Returns:
            None
        """
        _host: str
        if connection_settings is None:
            connection_settings = Connection()
        if isinstance(connection_settings, str):
            connection_settings = Path(connection_settings)
        if isinstance(connection_settings, Path):
            if not connection_settings.exists():
                raise FileNotFoundError(f"File {connection_settings} does not exist")
            with open(connection_settings, "r", encoding="utf-8") as _f:
                _settings = load(_f, Loader=SafeLoader)
                if not isinstance(_settings, dict):
                    raise TypeError(f"{_settings} is not a dict")
                _settings = cast(Dict[str, Any], _settings)
                if "rabbit" not in _settings:
                    raise ValueError(f"{_settings} does not contain 'rabbit'")
                _settings = cast(Dict[str, Any], _settings["rabbit"])
                if not isinstance(_settings, dict):
                    raise TypeError(f"{_settings} is not a dict")
                _settings = cast(Dict[str, Any], _settings)
                connection_settings = Connection.model_validate(_settings)

        _host = connection_settings.host

        if queue is None:
            # It should throw an exception, because I don't know which one to use
            queue = str(uuid4())

        _port: int = connection_settings.port

        _username: Union[str, None] = connection_settings.username
        _password: Union[str, None] = connection_settings.password
        _credentials: PlainCredentials
        if (
            _username is None
            or _username.strip() == ""
            or _password is None
            or _password.strip() == ""
        ):
            _credentials = ConnectionParameters.DEFAULT_CREDENTIALS
        else:
            _credentials = PlainCredentials(_username, _password)
        _retries: int = connection_settings.retries
        _retries = max(_retries, 1)
        _sleep_between: float = connection_settings.time_between_retries
        _sleep_between = max(_sleep_between, 0.1)
        for _i in range(_retries):
            try:
                self._client = BlockingConnection(
                    ConnectionParameters(
                        host=_host, port=_port, credentials=_credentials
                    )
                )
                break
            except AMQPConnectionError as _e:
                if (_i + 1) == _retries:
                    raise _e
                sleep(_sleep_between)
        self._queue = queue
        self._channel = self._client.channel()
        self.func = func
        self._data = {}
        self._input_type = input_type
        self._output_type = output_type

        self._on_finish = lambda x: None
        self._on_error = lambda x: None
        self._on_finish_signal = lambda: None

    @property
    def on_finish(self) -> Callable[[Union[U, List[U]]], None]:
        """Get a function to be called when the function is finished with a result"""
        return self._on_finish

    @on_finish.setter
    def on_finish(self, value: Callable[[Union[U, List[U]]], None]) -> None:
        """Set a function to be called when the function is finished with a result"""
        self._on_finish = value

    @property
    def on_error(self) -> Callable[[Exception], None]:
        """Get a function to be called when the function raises an exception"""
        return self._on_error

    @on_error.setter
    def on_error(self, value: Callable[[Exception], None]) -> None:
        """Set a function to be called when the function raises an exception"""
        self._on_error = value

    @property
    def on_finish_signal(self) -> Callable[[], None]:
        """Get a function to be called when a message is received"""
        return self._on_finish_signal

    @on_finish_signal.setter
    def on_finish_signal(self, value: Callable[[], None]) -> None:
        """Set a function to be called when a message is received"""
        self._on_finish_signal = value

    def run(self, what: Union[T, List[T], str]) -> Result[U]:
        """Send a message to the queue"""
        uuid = str(uuid4())
        self._data[uuid] = ResultModel(started=datetime.now())

        self._channel.queue_declare(queue=self._queue)
        self._channel.basic_publish(
            exchange="", routing_key=self._queue, body=dumps(what)
        )

        def _background_task(
            func: Callable[[Union[T, List[T], str]], Union[U, List[U]]],
        ) -> None:
            """Listen for messages in a separate thread"""
            self._listen(uuid, func)
            self._start()

        _th: Thread = Thread(
            target=_background_task,
            args=(self.func,),
            daemon=True,
        )
        _th.start()

        res: Result[U] = Result(
            uuid,
            self,
            self._on_finish,
            self._on_error,
            self._on_finish_signal,
        )
        res.connect()
        return res

    def _listen(
        self,
        uuid: str,
        callback: Callable[[Union[T, List[T], str]], Union[U, List[U]]],
    ) -> None:
        """Listen for messages in a separate thread"""
        self._channel.queue_declare(queue=self._queue)

        # pylint: disable=unused-argument,invalid-name
        def _internal_callback(a: Any, b: Any, c: Any, bingpot: AnyStr) -> None:
            """Listen for messages in a separate thread"""
            # pylint: disable=unused-argument
            try:
                if isinstance(bingpot, bytes):
                    bingpot = loads(bingpot)
                if isinstance(bingpot, str):
                    bingpot = json_loads(bingpot)
                if isinstance(bingpot, list):
                    bingpot = [
                        self._input_type.model_validate(i) if isinstance(i, dict) else i
                        for i in bingpot
                    ]
                if isinstance(bingpot, dict):
                    bingpot = self._input_type.model_validate(bingpot)
                self._data[uuid].result = callback(cast("Union[T, List[T]]", bingpot))
            except Exception as ex:  # pylint: disable=broad-except
                self._data[uuid].error = ex
            finally:
                self._data[uuid].completed = datetime.now()

        # pylint: enable=unused-argument,invalid-name
        self._channel.basic_consume(
            queue=self._queue, on_message_callback=_internal_callback, auto_ack=True
        )

    _th: Union[Thread, None] = None

    def _start(self, background: bool = True) -> None:
        """Start consuming messages."""
        if background:
            if self._th is None:

                self._th = Thread(target=self._channel.start_consuming)
                self._th.start()
        else:
            self._channel.start_consuming()

    def stop(self) -> None:
        """Stop consuming. This is called while exiting the context."""
        self._channel.stop_consuming()

    def delete_queue(self) -> None:
        """Delete the queue."""
        self._channel.queue_delete(self._queue)

    def close(self) -> None:
        """Close the connection and deletes the queue."""
        self.stop()
        self.delete_queue()
        self._client.close()

    def _is_connected(self) -> bool:
        """Check if the connection is open."""
        return bool(self._client.is_open)

    @property
    def client(self) -> BlockingConnection:
        """The RabbitMQ client.

        Returns:
            BlockingConnection: The pika client
        """
        return self._client

    @property
    def channel(self) -> BlockingChannel:
        """The BlockingChannel (if you want to do something fancy with it)

        Returns:
            BlockingChannel
        """
        return self._channel

    @property
    def is_connected(self) -> bool:
        """Check if the connection is open."""
        return self._is_connected()

    def __getitem__(self, key: str) -> ResultModel[U]:
        """Get the result."""
        return self._data[key]

    def __setitem__(self, key: str, value: ResultModel[U]) -> None:
        """Set the result."""
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        """Delete the result."""
        del self._data[key]


__all__ = (
    "RabbitBackgroundFunction",
    "T",
    "U",
)



"""Connection model"""

from typing import Union
from pydantic import BaseModel


class Connection(BaseModel):
    """Connection model"""

    # name: str
    host: str = "127.0.0.1"
    port: int = 5672
    username: Union[str, None] = None
    password: Union[str, None] = None
    retries: int = 20
    time_between_retries: float = 5.0


__all__ = ("Connection",)



"""Execution result"""

from contextlib import suppress
from typing import TYPE_CHECKING, Callable, Generic, TypeVar, Union, List, Any
from datetime import datetime
from threading import Thread
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)

if TYPE_CHECKING:  # pragma: no cover
    from .decorated import DecoratedBackgroundFunction
    from .rabbit import RabbitBackgroundFunction


class ExecutionResult(Generic[U]):
    """Execution result"""

    _dec: "Union[DecoratedBackgroundFunction[Any, U], RabbitBackgroundFunction[Any, U]]"
    _on_finish: "Callable[[Union[U, List[U]]], None]"
    _on_error: Callable[[Exception], None]
    _on_finish_signal: "Callable[[], None]"
    _started_thread: bool

    def __init__(  # noqa: D107 # pylint: disable=too-many-arguments
        self,
        uuid: str,
        dec: "Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]",
        on_finish: "Callable[[Union[U, List[U]]], None]",
        on_error: Callable[[Exception], None],
        on_finish_signal: "Callable[[], None]",
    ) -> None:
        self._uuid = uuid
        self._dec = dec
        self._on_finish = on_finish
        self._on_error = on_error
        self._on_finish_signal = on_finish_signal
        self._started_thread = False

    _uuid: str

    @property
    def uuid(self) -> str:
        """The UUID of the execution result"""
        return self._uuid

    @property
    def result(self) -> Union[U, List[U], None]:
        """The result of the execution"""
        return self._dec[self.uuid].result

    @property
    def started(self) -> Union[datetime, None]:
        """The start time of the execution"""
        return self._dec[self.uuid].started

    @property
    def completed(self) -> Union[datetime, None]:
        """The completion time of the execution"""
        return self._dec[self.uuid].completed

    @property
    def finished(self) -> bool:
        """Whether the execution is finished"""
        return self._dec[self.uuid].completed is not None

    @property
    def error(self) -> Union[Exception, None]:
        """The error of the execution"""
        return self._dec[self.uuid].error

    @property
    def success(self) -> bool:
        """Whether the execution was successful"""
        return self._dec[self.uuid].error is None and self.finished

    def _check(
        self,
        timeout: Union[float, None] = None,  # pylint: disable=unused-argument
    ) -> None:
        """
        Check the execution result and perform the necessary actions based on the result status.

        This method continuously checks the status of the execution result. If an error is \
              present, it calls the `_on_error` callback function with the error as an \
              argument and returns. If the execution is finished, it retrieves the \
              result and checks if it is None. If it is None, it calls \
              the `_on_error` callback function with a `ValueError` indicating that\
              the result is None and returns. If the result is not None, it calls\
              the `_on_finish` callback function with the result as an argument, \
              calls the `_on_finish_signal` callback function, and returns.

        Parameters:
            timeout (Union[float, None], optional): The maximum time in seconds\
                  to wait for the execution to complete. Defaults to None.

        Returns:
            None
        """
        # TODO: timeout
        while True:
            if self.error is not None:
                self._on_error(self.error)
                return
            if self.finished:
                data = self.result
                if data is None:
                    self._on_error(ValueError("Result is None"))
                    return
                self._on_finish(data)
                self._on_finish_signal()
                return

    def connect(self) -> None:
        """Connect to the background task"""
        if self._started_thread:
            return
        t = Thread(target=self._check)
        t.start()
        self._started_thread = True

    def __del__(self) -> None:
        """Delete the result."""
        with suppress(KeyError):
            del self._dec[self.uuid]


__all__ = ("ExecutionResult", "T", "U")



"""The Coleridge class"""

from pathlib import Path
from typing import Literal, Callable, Union, List, cast
from .decorator import ColeridgeDecorator, T, U
from .decorated import DecoratedBackgroundFunction
from .models.connection import Connection
from .rabbit import RabbitBackgroundFunction
from .get_types import get_params_type


class Coleridge:
    """The Coleridge class is a decorator that allows you to easily \
    create decorators that can be used to decorate functions with RabbitMQ \
    or background tasks.
    """

    _connection_settings: Union[Connection, None, str, Path]
    _queue: Union[str, None]
    _mode: Literal["rabbit", "background"]

    def __init__(
        self,
        connection_settings: Union[Connection, None, str, Path] = None,
        queue: Union[str, None] = None,
        mode: Literal["rabbit", "background"] = "background",
    ) -> None:
        """
        Initializes a Coleridge object.

        Args:
            connection_settings (Union[Connection, None, str, Path]): The connection \
                settings for the Coleridge object. Defaults to None.
            queue (Union[str, None]): The queue for the Coleridge object. Defaults to None.
            mode (Literal["rabbit", "background"]): The mode for the Coleridge object. \
              Defaults to "background".

        Returns:
            None
        """
        self._connection_settings = connection_settings
        self._queue = queue
        self._mode = mode

    def magic_decorator(
        self,
        queue: Union[str, None] = None,
        on_finish: Union[Callable[[Union[U, List[U]]], None], None] = None,
        on_error: Union[Callable[[Exception], None], None] = None,
        on_finish_signal: Union[Callable[[], None], None] = None,
    ) -> Callable[
        [Callable[[Union[T, List[T]]], Union[U, List[U]]]],
        Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]],
    ]:
        """
        A decorator function that creates a ColeridgeDecorator instance and \
        returns a decorated function.
        
        Parameters:
        queue (Union[str, None]): The queue to be used for the decorated function.
        on_finish (Union[Callable[[Union[U, List[U]]], None], None]): The callback \
        function to be executed when the decorated function finishes.
        on_error (Union[Callable[[Exception], None], None]): The callback function to be \
            executed when the decorated function encounters an error.
        on_finish_signal (Union[Callable[[], None], None]): The callback function to be \
        executed when the decorated function finishes with a signal.
        
        Returns:
        Callable[[Callable[[Union[T, List[T]]], Union[U, List[U]]]], \
            Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]]: \
                A decorator function that takes a function as input and returns a \
                    decorated function.
        """

        def _inner(
            func: Callable[[Union[T, List[T]]], Union[U, List[U]]]
        ) -> Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]:
            """
            Inner function of the magic decorator, responsible for creating a \
                ColeridgeDecorator instance.
            
            Parameters:
            func (Callable[[Union[T, List[T]]], Union[U, List[U]]]): The function to be decorated.
            
            Returns:
            Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]: The \
            decorated function.
            """
            input_type, output_type = get_params_type(func)
            dec = ColeridgeDecorator(
                input_type,
                output_type,
                mode=self._mode,
                connection_settings=self._connection_settings,
                queue=self._queue if queue is None else queue,
                on_finish=on_finish,
                on_error=on_error,
                on_finish_signal=on_finish_signal,
            )
            return dec(func)

        return _inner

    def __call__(
        self,
        func: Callable[[Union[T, List[T]]], Union[U, List[U]]],
    ) -> Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]:
        """
        Calls the magic decorator function with the provided function as an argument.

        Parameters:
        func (Callable[[Union[T, List[T]]], Union[U, List[U]]]): The function to be decorated.

        Returns:
        Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]]: \
            The decorated function.
        """
        return cast(
            Union[DecoratedBackgroundFunction[T, U], RabbitBackgroundFunction[T, U]],
            self.magic_decorator()(
                func,  # type: ignore[arg-type]
            ),
        )


__all__ = ("Coleridge",)



"""Response model"""

from datetime import datetime
from typing import Union, TypeVar, Generic, List
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class ResultModel(BaseModel, Generic[T]):
    """Response model"""

    result: Union[T, List[T], None] = None
    started: Union[datetime, None] = None
    completed: Union[datetime, None] = None
    error: Union[Exception, None] = None

    class Config:  # pylint: disable=too-few-public-methods
        """Pydantic config"""

        arbitrary_types_allowed = True


__all__ = ("ResultModel",)



"""Decorated background function"""

from datetime import datetime
from uuid import uuid4
from threading import Thread
from json import loads
from typing import TypeVar, Generic, Callable, Union, List, Dict, Type, cast
from pydantic import BaseModel
from .models.response import ResultModel
from .result import ExecutionResult as Result

T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)


class DecoratedBackgroundFunction(Generic[T, U]):
    """Decorated background function"""

    _data: Dict[str, ResultModel[U]]
    _input_type: Type[T]
    _output_type: Type[U]
    _on_finish: Callable[[Union[U, List[U]]], None]
    _on_error: Callable[[Exception], None]
    _on_finish_signal: Callable[[], None]
    func: Callable[[Union[T, List[T]]], Union[U, List[U]]]

    def __init__(
        self,
        func: Callable[[Union[T, List[T]]], Union[U, List[U]]],
        input_type: Type[T],
        output_type: Type[U],
    ) -> None:
        """
        Initializes a new instance of the DecoratedBackgroundFunction class.

        Args:
            func: A callable function that takes a single argument of type T or a list of T \
                and returns a single value of type U or a list of U.
            input_type: The type of the input argument.
            output_type: The type of the output value.

        Returns:
            None
        """
        self.func = func
        self._data = {}
        self._input_type = input_type
        self._output_type = output_type

        self._on_finish = lambda x: None
        self._on_error = lambda x: None
        self._on_finish_signal = lambda: None

    @property
    def on_finish(self) -> Callable[[Union[U, List[U]]], None]:
        """Get a function to be called when the function is finished with a result"""
        return self._on_finish

    @on_finish.setter
    def on_finish(self, value: Callable[[Union[U, List[U]]], None]) -> None:
        """Set a function to be called when the function is finished with a result"""
        self._on_finish = value

    @property
    def on_error(self) -> Callable[[Exception], None]:
        """Get a function to be called when the function raises an exception"""
        return self._on_error

    @on_error.setter
    def on_error(self, value: Callable[[Exception], None]) -> None:
        """Set a function to be called when the function raises an exception"""
        self._on_error = value

    @property
    def on_finish_signal(self) -> Callable[[], None]:
        """Get a function to be called when a message is received"""
        return self._on_finish_signal

    @on_finish_signal.setter
    def on_finish_signal(self, value: Callable[[], None]) -> None:
        """Set a function to be called when a message is received"""
        self._on_finish_signal = value

    def _run_background(
        self,
        input_value: Union[T, List[T], str],
        uuid: str,
    ) -> None:
        """
        Runs the decorated function in the background with the provided input value and uuid.

        Args:
            input_value: The input value to be passed to the decorated function. It can be a \
                  string, a list, or a dictionary.
            uuid: A unique identifier for the background task.

        Returns:
            None
        """
        try:
            if isinstance(input_value, str):
                input_value = loads(input_value)
            if isinstance(input_value, list):
                input_value = [
                    self._input_type.model_validate(i) if isinstance(i, dict) else i
                    for i in input_value
                ]
            if isinstance(input_value, dict):
                # pylint: disable=line-too-long
                input_value = self._input_type.model_validate(input_value)  # type: ignore[unreachable]
                # pylint: enable=line-too-long
            self._data[uuid].result = self.func(cast("Union[T, List[T]]", input_value))
        except Exception as ex:  # pylint: disable=broad-except
            self._data[uuid].error = ex
        finally:
            self._data[uuid].completed = datetime.now()

    def run(  # noqa: D102
        self,
        input_value: Union[T, List[T], str],
        timeout: Union[float, None] = None,  # pylint: disable=unused-argument
    ) -> Result[U]:
        """
        Run a background task with the given input value and optional timeout.

        Args:
            input_value: The input value to be processed by the background task.
            timeout: The maximum time in seconds to wait for the task to complete.

        Returns:
            A Result object representing the outcome of the background task.
        """
        uuid = str(uuid4())
        self._data[uuid] = ResultModel(started=datetime.now())

        t = Thread(target=self._run_background, args=(input_value, uuid))
        t.start()
        res: Result[U] = Result(
            uuid,
            self,
            self._on_finish,
            self._on_error,
            self._on_finish_signal,
        )
        res.connect()
        return res

    def __getitem__(self, key: str) -> ResultModel[U]:
        """Get the result."""
        return self._data[key]

    def __setitem__(self, key: str, value: ResultModel[U]) -> None:
        """Set the result."""
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        """Delete the result."""
        del self._data[key]



"""Get the types of the parameters and return values of a function."""

from typing import Tuple, Callable, Type, Union, List, TypeVar
from inspect import signature
from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)
U = TypeVar("U", bound=BaseModel)


def get_params_type(
    func: Callable[[Union[T, List[T]]], Union[U, List[U]]]
) -> Tuple[Type[T], Type[U]]:
    """Get the types of the parameters and return values of a function."""
    sig = signature(func)
    key = list(sig.parameters.keys())[0]
    first_param_type = sig.parameters[key].annotation
    return_type = sig.return_annotation
    if first_param_type is Union:
        first_param_type = first_param_type.__args__[0]
    elif first_param_type is List:
        first_param_type = first_param_type.__args__[0]
    if return_type is Union:
        return_type = return_type.__args__[0]
    elif return_type is List:
        return_type = return_type.__args__[0]
    if not issubclass(first_param_type, BaseModel):
        raise TypeError(f"{first_param_type} is not a BaseModel")
    if not issubclass(return_type, BaseModel):
        raise TypeError(f"{return_type} is not a BaseModel")
    return first_param_type, return_type


__all__ = ("get_params_type",)
